/*
 *	Functions for analysing i386 architecture systems
*/

#include "../../globals.h"
#include "magic.h"

/*
 * Some alchemy:
 */

/*
 * Function that manages mapping from linear to physical address space.
 * 
 * priv_block_size from lin_addr are read using provided PGD 
 *(kernel PGD by default)
 * into out_data pointer.
 * dump_offset is output parameter, if NULL, ignored.
 * Otherwise physical offset translated from linear address is stored here.
 * (!) Please note, that read block can be larger than one page. 
 * In this case offset is valid only for first part of data.
 */

int i386_read_linear(u_int64_t PGD, u_int64_t lin_addr, void *out_data,
			 u_int64_t priv_block_size, u_int64_t *dump_offset){

	// "PGD" is needed, because each process have 
	// it's own linear address space.
	// Process 0 means kernel memory.

	// Find in VPT "physical" address, where data should be. 
	// (if problem, return 1)
	// Read block_size of that data into out_data 
	// (if problem, return 2)
	// return 0

	u_int64_t physical_offset = lin_addr;
	u_int64_t PDE_offset = lin_addr;
	u_int64_t PTE_offset = lin_addr;
	u_int64_t page_number = lin_addr;
	u_int64_t tmp_var = 0;

	if (PGD == 0) {	//special case
		PGD = img1.VPT_address;

	/* if VPT_address is not available,
	 * we can try "best effort", and use hard-coded 
	 * "direct mapping", usually used by OS kernels.
	 * (and OS specific)
	 * for linux on i386, 0xC0000000 can be subtracted 
	 * from linear kernelspace addresses.
	 * Does not work for BSD (!)
  	 */  
		if (img1.VPT_not_available == 1) { 
			// kernel text direct mapping
			if ((lin_addr > 0xC0000000) && 
			    (lin_addr < 0xFFFFFFFF)) {
				physical_offset=lin_addr - 0xC0000000;
				dbgprint("direct mapping 0x%llX -> 0x%llX",\
					lin_addr,physical_offset);
			}
			
			// If none of mappings matches 
			if (physical_offset == lin_addr) {
				dbgprint("nothing matched, 0x%llX", lin_addr);
				return -1;
			}
			
			// Read memory

 			// level 0 (physical page) 
			lseek(inputfd,physical_offset,SEEK_SET);
			dbgprint("(level 0)Reading from PO: 0x%llX\n",
				physical_offset);
			
			if (dump_offset != NULL) { 
				// save physical offset to output variable
				*dump_offset=physical_offset; 
			}

			status = read(inputfd, out_data, priv_block_size);
			if ( status == -1 ) {
        		        errormsg=strerror(errno);
               			fprintf(stderr,
					"read_liner: read() in %s: %s.\n",
					inputfile,errormsg);

	                	return 38;
			}

			return 0;
		}

	} // else PGD is linear address of PGD of some process

	dbgprint("PGD1: %llX, addr: %llX",PGD,lin_addr);

	//Kernel linear address space:

	// For extracting addresses I use bitwise and,
	// for extracting offsets I have to shift bits.

	// 4-Mbyte Page Translation

	// bits 21:0  
        physical_offset = physical_offset & B64(0,21);
	// bits 31:22 
        PDE_offset = BS64(PDE_offset,22,31);

	dbgprint("lin_addr: %llX, Phys_Off: 0x%llX, PDE_Off: 0x%llX, "\
		"PDE_e: 0x%llX\n", lin_addr, physical_offset, PDE_offset,
		PGD+PDE_offset*img1.pointer_size); 

	// *4 because each entry have 32 bits = 4bytes
        lseek(inputfd,PGD+PDE_offset*img1.pointer_size,SEEK_SET);
		
        if ( (status=read(inputfd,&page_number,img1.pointer_size)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "read_liner: read() in %s: %s.\n",
			inputfile, errormsg);
                return 35;
        }

	dbgprint("page_number: 0x%llX, pointer size: %d\n",
		page_number,img1.pointer_size);

	// TODO: process all necessary special bit, like swapped, etc...

	tmp_var = page_number;
	tmp_var = tmp_var & B64(7,7); //only bit NUMBER 7 (ie. 8th bit)

	if (tmp_var != 0) { //4MB pages are used
		
		dbgprint("4MB pages used for this page.\n"); 

		//bits 31:22
		page_number = BS64(page_number,22,31); 

		// If we want to read more than 1 page:
		// (use recursion;)
		//
		if (priv_block_size > (MB_4-physical_offset)){ 
			// offset + block_size must be less then 4M 
			dbgprint("page too small for reading..."\
				 "processing another one.\n");

			// recursion
			tmp_var = MB_4-physical_offset;
			i386_read_linear(PGD, lin_addr+tmp_var,
					out_data+tmp_var,
					priv_block_size-tmp_var, NULL);

		} else{
			tmp_var=priv_block_size;
		}	

 		//(4MB page number "large_page_number" + offset)
		lseek(inputfd,page_number*MB_4+physical_offset,SEEK_SET);
		dbgprint("4MB page_number: 0x%llX, phys. addr: 0x%llX\n",
			page_number,page_number*MB_4+physical_offset);

		if (dump_offset != NULL) { 
			//save physical offset to output variable
			*dump_offset=page_number*MB_4+physical_offset;
		}
		
		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        	        errormsg=strerror(errno);
                	fprintf(stderr, "i386_read_linear: read() in %s: %s.\n",\
				inputfile,errormsg);
	                return 35;
        	}
	} else {   //4KB pages are used
		dbgprint("4KB pages used for this page.\n");
	
		page_number = page_number & B64(12,31); //bits 31:12 
		PTE_offset = BS64(lin_addr,12,21); //bits 21:12
		
		physical_offset = physical_offset & B64(0,11); //bits 11:0
		
		dbgprint("4KB Searching for PTE in 0x%llx.\n",page_number);

		lseek(inputfd, page_number+PTE_offset*img1.pointer_size,
			SEEK_SET); 
		
		status=read(inputfd,&tmp_var,img1.pointer_size);
		if ( status == -1 ) {
                 	errormsg=strerror(errno);
		        fprintf(stderr, "read_liner: read() in %s: %s.\n",
				inputfile,errormsg);
			return 36;
		}
		
		dbgprint("4KB In PTE I found 0x%llX.\n",tmp_var);

		// bits 31:12 //PTE entry
		tmp_var = tmp_var & B64(12,31); 

		dbgprint("4KB Seeking to address 0x%llX.\n",
			tmp_var+physical_offset);

		lseek(inputfd,tmp_var+physical_offset,SEEK_SET);

		if (dump_offset != NULL) { 
			//save physical offset to output variable
			*dump_offset=tmp_var+physical_offset;
		}

		// If we want to read more than 1 page:
		// (use recursion;)
		//
		if (priv_block_size > (KB_4-physical_offset)){ 
			// offset + block_size must be less then 4M
			dbgprint("page too small for reading... "\
				"processing another one.\n");

			// recursion
			tmp_var = KB_4-physical_offset;
			// Process rest of request recursively... 
			// beware about effectiveness
			i386_read_linear(PGD, lin_addr+tmp_var,
					out_data+tmp_var,
					priv_block_size-tmp_var, NULL);

		} else{
			tmp_var=priv_block_size;
		}

		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        	        errormsg=strerror(errno);
                	fprintf(stderr, "read_liner: read() in %s: %s.\n",
				inputfile,errormsg);
	                return 38;
        	}


	}

	return 0;

}


/*
 * Function for verifying if given address contains system VPT 
 * (resp. PaGeDirectory)
 *
 * Verification is done by checking:
 * - If first I386_VPT_MAGIC_ZEROES bytes contain only zeroes.
 *   this is linux-based heuristic, but seems to work well
 *   on different OS

 * This should usually be true, but it is just heuristic, 
 * and can be wrong.
 *
 * i386_VPT_test (physical address to verify )
 * return 0 on success, otherwise error > 0
 *
 * */

int i386_VPT_test(u_int64_t my_addr){

	int mi=0,ecounter=0; //error counter
	unsigned char small_page[I386_VPT_MAGIC_BS];

	dbgprint("trying  0x%llX.\n",my_addr); 

	lseek(inputfd,my_addr,SEEK_SET); 
	if ( (status=read(inputfd,&small_page,I386_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_VPT_test: read() in %s: %s.\n",
			inputfile, errormsg);
                return 99;
        }

	// Addresses below 0xC0000000 should contain mostly zeroes for
	// kernel PGD. This heuristic is based on linux, but 0xC0000000
	// is used also on bsd, and "seems" to work for windows too.
	// This definitely can fail.
	
	for (mi=0;mi < I386_VPT_MAGIC_ZEROES;mi++) { //first 
		if (small_page[mi] != 0) {
			if (++ecounter > I386_VPT_MAGIC_ZEROES_ERR) {
				dbgprint("addr: 0x%llX mi=%d, ecounter: %d\n",
					my_addr,mi,ecounter);
				return 1; //tolerate small amount of errors
			}
		}
	}

	// check if structure *SEEMS* to be regular.
	// (!) NOTE: This is just simple heuristic, success not guaranteed
	// see documentation for more information

	// This is first last byte of record 
	// (in fact first, because of little endian)
	// In example bellow it is "E3"
	// Bits   MBZ IGN A PCD PWT U/S R/W  P      (for 4KB PDE)
	// 0x6b ~  0   1  1  0   1   0   1   1 

	// Typical PGD have entries like this (example):
	// E3 01 00 01  
	// E3 01 40 01  
	// E3 01 80 01  
	// E3 01 C0 01 
	// E3 01 00 02

	ecounter=0;

	for (; mi < (I386_VPT_MAGIC_ZEROES + I386_VPT_MAGIC_ENTRIES_CHECK);
		mi += img1.pointer_size) { 

		// test if nearby entries are same, 
		// (only some bits of first byte from each pointer)
		// and avoid trivial cases (0x0, 0x7F)
		if (((small_page[mi] & 0x6b) != \
		    (small_page[mi+img1.pointer_size] & 0x6b))\
		    ||(small_page[mi] == 0)\
		    ||(small_page[mi] == 0x7F)) {
			if (++ecounter > I386_VPT_MAGIC_ENTRIES_ERR) {
				dbgprint("addr: 0x%llX ,entries, mi=%d\n",
					my_addr,mi); 
				return 2; //tolerate small amount of errors
			}
		}
	}
	dbgprint("addr: 0x%llX ,entries, mi=%d, err=%d\n",my_addr,mi,ecounter);

	return 0; //all OK			
}

#if 0

/* An attemp to make "cleaner" implementation of i386_VPT_test.
 * Unfortunately, there are too many false positives,
 * So I decided to keep previou dirty "magic" function,
 * even thought it is clear it can fail sometimes.
 * Overal result of older function are very good.
 * 
 */

/*
 * Function for verifying if given address contains system VPT 
 * (resp. PaGeDirectory)
 *
 * i386_VPT_test (physical address to verify )
 * return 0 on success, otherwise error > 0
 *
 * */

int i386_VPT_test(u_int64_t my_addr){

	u_int64_t tmp_var;
	int mi,ecounter=0,rcounter=0;

	// using 32bit variables. (!)
	u_int32_t small_page[I386_VPT_MAGIC_BS/I386_POINTER_SIZE];

	// must be 4KB aligned

	tmp_var = my_addr & 0x0000000000000FFFLLU;
	
	dbgprint("0x%llX",my_addr);

	if (tmp_var != 0) {
		dbgprint("Not 4KB aligned!");
		return 1;
	}

	lseek(inputfd,my_addr,SEEK_SET); 

	// read whole page (4KB)
	if ( (status=read(inputfd,&small_page,I386_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	for (mi=0;mi < (I386_VPT_MAGIC_BS/I386_POINTER_SIZE);mi++) { //page
	
		if (small_page[mi] != 0) { //if not empty record
			// If PAE is used, we are reading PDPE
			// Then, bits 8,7,6,5,2 and 1 MustBeZero
			
			// If NON-PAE is used then:
			// If bit 7 is 1,
			// 4MB pages are used, and bit 21 MustBeZero
			// If bit 7 is 0, 
			// 4KB pages are used, and there should be 
			// pointer to PageTable

			rcounter++;
			if ((small_page[mi] & 0x00000080LU) == 1) { //bit nr. 7
				// If bit 7 is 1,
				// 4MB pages are used, and bit 21 MustBeZero
				dbgprint("4MB entry");
				if ((small_page[mi] & 0x00000080LU) == 1) { //bit nr. 7
					if (ecounter++ > I386_VPT_MAGIC_ERR_FATAL) {
						dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",\
							my_addr,mi,ecounter,rcounter);
						return 2; //tolerate small amount of errors
					}
				}
			} else {
				//NEXT TODO
				// Assume PAE is not working;)
				// So this is 4KB NON-PAE page.
							

			}
		}	

	}	


	if (rcounter < I386_VPT_MIN_REC) {
		dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",\
			my_addr,mi,ecounter,rcounter);
		return 4; // require at least few valid records
	}

	// Additional check: 
	// errors/valid records must pass some ratio
	// Direct seeking for errors is too strict for full tables

	if (ecounter != 0) { //division by zero
		if ( (rcounter/ecounter) < I386_VPT_MAGIC_ERR_RATIO) {
			dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d",\
				my_addr,mi,ecounter,rcounter);
			return 5; // require low error ratio
		}
	}

	return 0; //all OK			
}

#endif

/*
 * Function for verifying if given address contains VPT 
 * (resp. PaGeDirectory) of some process.
 *
 * Verification is done by simple checking
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, as userspace PGD are 
 * cloned from kernel PGD, but it is just heuristic, 
 * and can be wrong.
 * 
 *  XXX rewrite
 *
 *  i386_user_I386_VPT_test (linear address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */


int i386_user_VPT_test(u_int64_t my_addr){

	int mi=0;
	unsigned char small_page[I386_VPT_MAGIC_BS];

	if ( (status=i386_read_linear(0,my_addr,small_page,
		I386_VPT_MAGIC_BS,NULL)) != 0 ) {

                errormsg=strerror(errno);
                fprintf(stderr, "i386_user_I386_VPT_test: read() in %s: %s.\n",\
			inputfile,errormsg);
                return 99;
        }
	//XXX CHANGE
	
	dbgprint("%llx\n",my_addr);

	mi=I386_VPT_MAGIC_ZEROES; 

	//check if structure seems to be regular.
	// (!) NOTE: This is just simple heuristic, success not guaranteed
	//see documentation for more information
	// TODO: Change this to same test as used for kernel VPT?



	if ((small_page[mi] & 0x6f) != (small_page[mi+img1.pointer_size] & 0x6f)) 
		{return 2;}
	if ((small_page[mi] & 0x6f) != (small_page[mi+img1.pointer_size*2] & 0x6f)) 
		{return 3;}
	if (small_page[mi] == 0) 
		{return 4;}

	dbgprint("dbg2: :%d,%d,%c\n", small_page[mi], 
		small_page[mi], small_page[mi]);

	return 0; //all OK			
}

/*
 * Function that tries to determine general information 
 * about analyzed image.
 *
 * for now, only system VPT is found.
 * if mode == 1, then all possible VPT candidates are printed.
 *
 * i386_find_VPT(mode)
 * return 0 on success, otherwise error > 0
 * Discovered information are saved in global structure img1.
 *
 * */

int i386_find_VPT(int mode){
	
	u_int64_t addr=I386_VPT_MAGIC_ADDR;
	
	// I realised, that I don't get too much information in 
	// this phase of analysis 

	// Special mode, only find all possible VPT candidates
	if (mode == 1) {
		printf("Listing VPT candidates(i386):\n");
		while ( (addr + I386_VPT_MAGIC_BS) <= img1.ram_size) { 
			if (i386_VPT_test(addr) == 0) {
				printf("	Possible VPT at: 0x%llX\n",
					addr);
			}
			addr += I386_VPT_MAGIC_BS; // 4KB
			dbgprint("step, addr: 0x%llX\n",addr);
		}

		return 0;	
	} 

	// Otherwise normal approach

	printf("i386_find_VPT:\n");
	if (img1.VPT_user_specified == 0 ) {	
		// if image address was not manually specified
		while (i386_VPT_test(addr) != 0) {
	
			addr += I386_VPT_MAGIC_BS; // 4KB
			dbgprint("step, addr: 0x%llX\n",addr);

			if (addr > ( I386_VPT_MAGIC_ADDR + 
				(I386_VPT_MAGIC_BS * I386_VPT_MAGIC_COUNT))) {
				// we have tried too many pages, 
				// so something is wrong probably
				// further search will be useless
				printf("I386_VPT_address not found, \
					terminating at: 0x%llX\n",addr);
				return 1;
			}
		}
	
	
		img1.VPT_address=addr; // found one! :)
		printf("VPT address found at: 0x%llX\n",img1.VPT_address);
	} else {
		if (img1.VPT_not_available == 0 ) {	
			// Use of PGD is not disabled	
			printf("User specified VPT address: 0x%llX\n",
				img1.VPT_address);
		} else {
			printf("Using hardcoded mappings instead of VPT. "\
				"This may produce strange results!\n");
		}
	}
	
	return 0;  // 0 if all OK, otherwise error
}

/*
 * Function that sets pointers to all other 
 * architecture dependent functions and hooks
 *
 * arm_arch_settings()
 * return 0 on success 
 */

int i386_arch_settings(){

	strncpy(img1.architecture,"i386\0",5);
	img1.read_linear=(int (*)())i386_read_linear;
	img1.VPT_test=(int (*)())i386_VPT_test;
	img1.user_VPT_test=(int (*)())i386_user_VPT_test;
	img1.find_VPT=(int (*)())i386_find_VPT;
	img1.pointer_size=I386_POINTER_SIZE;
	img1.integer_size=I386_INTEGER_SIZE;

	return 0;
}

int i386_init(struct arch_struct *self){

	strncpy(self->architecture,"i386\0",5);
	self->read_linear=(int (*)())i386_read_linear;
	self->VPT_test=(int (*)())i386_VPT_test;
	self->user_VPT_test=(int (*)())i386_user_VPT_test;
	self->find_VPT=(int (*)())i386_find_VPT;
	self->pointer_size=I386_POINTER_SIZE;
	self->integer_size=I386_INTEGER_SIZE;

	return 0;
}



